from typing import Any, Dict, List, Literal

from ....action import Actions, CreateDiceAction, CreateObjectAction, RemoveObjectAction
from .....utils.class_registry import register_class
from ....status.character_status.base import (
    DefendCharacterStatus,
    ReviveCharacterStatus,
    RoundEndAttackCharacterStatus,
)
from .....utils import BaseModel
from ....event import (
    CreateObjectEventArguments,
    MakeDamageEventArguments,
    PlayerActionStartEventArguments,
    RemoveObjectEventArguments,
    RoundPrepareEventArguments,
)
from ....match import Match
from ....struct import Cost, ObjectPosition
from ....consts import (
    DAMAGE_TYPE_TO_ELEMENT,
    ELEMENT_TO_DIE_COLOR,
    CostLabels,
    DamageElementalType,
    DieColor,
    ElementType,
    FactionType,
    IconType,
    ObjectPositionType,
    WeaponType,
)
from ....character.character_base import (
    CharacterBase,
    CreateStatusPassiveSkill,
    ElementalBurstBase,
    ElementalNormalAttackBase,
    ElementalSkillBase,
    TalentBase,
)
from .....utils.desc_registry import DescDictType


class PainForPainStatus_4_3(DefendCharacterStatus):
    name: Literal["Pain for Pain"] = "Pain for Pain"
    version: Literal["4.3"] = "4.3"
    usage: int = 1
    max_usage: int = 1
    min_damage_to_trigger: int = 3
    max_in_one_time: int = 1

    def event_handler_MAKE_DAMAGE(
        self, event: MakeDamageEventArguments, match: Match
    ) -> List[RemoveObjectAction | CreateObjectAction]:
        """
        After make damage, if self usage becomes 0 (which means it negates
        DMG), super will remove self. At this time, create status for opposite
        active character.
        """
        ret: List[RemoveObjectAction | CreateObjectAction] = []
        ret += super().event_handler_MAKE_DAMAGE(event, match)
        if len(ret) == 0:
            # not remove self, do nothing
            return ret
        target = match.player_tables[
            1 - self.position.player_idx
        ].get_active_character()
        our = match.player_tables[self.position.player_idx].characters[
            self.position.character_idx
        ]
        assert our.name == "Signora"
        if our.element == ElementType.CRYO:
            status_name = "Sheer Cold"
        elif our.element == ElementType.PYRO:
            status_name = "Blazing Heat"
        else:
            raise AssertionError(f"Unknown element {our.element}")
        ret.append(
            CreateObjectAction(
                object_position=target.position.set_area(
                    ObjectPositionType.CHARACTER_STATUS
                ),
                object_name=status_name,
                object_arguments={},
            )
        )
        return ret


class IceSealedCrimsonWitchOfEmbers_4_3(ReviveCharacterStatus):
    name: Literal[
        "Ice-Sealed Crimson Witch of Embers"
    ] = "Ice-Sealed Crimson Witch of Embers"
    version: Literal["4.3"] = "4.3"
    heal: int = 1
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS

    def event_handler_ROUND_PREPARE(
        self, event: PlayerActionStartEventArguments, match: Match
    ) -> List[RemoveObjectAction]:
        character = match.player_tables[self.position.player_idx].characters[
            self.position.character_idx
        ]
        if character.hp > 4:
            # hp > 4, not remove
            return []
        # hp <= 4, remove
        return [RemoveObjectAction(object_position=self.position)]


class SheerCold_4_3(RoundEndAttackCharacterStatus):
    name: Literal[
        "Sheer Cold",
        "Blazing Heat",
    ]
    version: Literal["4.3"] = "4.3"
    usage: int = 1
    max_usage: int = 1
    damage: int = 1

    # will modify when init
    damage_elemental_type: DamageElementalType = DamageElementalType.PIERCING
    icon_type: Literal[
        IconType.DEBUFF_ELEMENT_ICE, IconType.DEBUFF_ELEMENT_FIRE
    ] = IconType.DEBUFF_ELEMENT_ICE

    def __init__(self, *argv, **kwargs):
        super().__init__(*argv, **kwargs)
        if self.name == "Sheer Cold":
            self.damage_elemental_type = DamageElementalType.CRYO
            self.icon_type = IconType.DEBUFF_ELEMENT_ICE
        elif self.name == "Blazing Heat":
            self.damage_elemental_type = DamageElementalType.PYRO
            self.icon_type = IconType.DEBUFF_ELEMENT_FIRE
        else:
            raise AssertionError(f"Unknown name {self.name}")

    def event_handler_CREATE_OBJECT(
        self, event: CreateObjectEventArguments, match: Match
    ) -> List[RemoveObjectAction]:
        """
        When a character attaches two status, remove status in previous.
        This will only perform on Blazing Heat to avoid generating multiple
        RemoveObjectAction for same status.
        """
        if self.name != "Blazing Heat":
            # not Blazing Heat, do nothing
            return []
        status = (
            match.player_tables[self.position.player_idx]
            .characters[self.position.character_idx]
            .status
        )
        blazing_heat = None
        sheer_cold = None
        for sidx, s in enumerate(status):
            if s.name == "Blazing Heat":
                blazing_heat = [sidx, s]
            elif s.name == "Sheer Cold":
                sheer_cold = [sidx, s]
        assert blazing_heat is not None
        if sheer_cold is None:
            # no sheer cold, do nothing
            return []
        # remove status in previous
        assert blazing_heat[0] != sheer_cold[0]
        if blazing_heat[0] < sheer_cold[0]:
            target = blazing_heat[1]
        else:
            target = sheer_cold[1]
        return [
            RemoveObjectAction(object_position=target.position),
        ]


class SignoraSkillValidCheck(BaseModel):
    """
    Used in all Signora's skills to check if skill is valid.
    """

    position: ObjectPosition = ObjectPosition(
        player_idx=-1,
        area=ObjectPositionType.INVALID,
        id=-1,
    )
    damage_type: DamageElementalType

    def is_valid(self, match: Match) -> bool:
        character = match.player_tables[self.position.player_idx].characters[
            self.position.character_idx
        ]
        assert character.name == "Signora"
        return (character.desc == "") == (self.damage_type == DamageElementalType.CRYO)


class SignoraNormalAttack(SignoraSkillValidCheck, ElementalNormalAttackBase):
    """
    availability based on Signora's desc.
    """

    name: Literal[
        "Frostblade Hailstorm",
        "Fire Normal Attack",
    ]

    # will modify when init
    damage_type: DamageElementalType = DamageElementalType.PIERCING
    cost: Cost = Cost()

    def __init__(self, *argv, **kwargs):
        super().__init__(*argv, **kwargs)
        if self.name == "Frostblade Hailstorm":
            self.damage_type = DamageElementalType.CRYO
        elif self.name == "Fire Normal Attack":
            self.damage_type = DamageElementalType.PYRO
        else:
            raise AssertionError(f"Unknown name {self.name}")
        self.cost = Cost(
            elemental_dice_color=ELEMENT_TO_DIE_COLOR[
                DAMAGE_TYPE_TO_ELEMENT[self.damage_type]
            ],
            elemental_dice_number=1,
            any_dice_number=2,
        )


class SignoraElementalSkill(SignoraSkillValidCheck, ElementalSkillBase):
    """
    availability based on Signora's desc.
    """

    name: Literal[
        "Biting Shards",
        "Fire Elemental Skill",
    ]

    # will modify when init
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.PIERCING
    cost: Cost = Cost(elemental_dice_number=3)

    def __init__(self, *argv, **kwargs):
        super().__init__(*argv, **kwargs)
        if self.name == "Biting Shards":
            self.damage_type = DamageElementalType.CRYO
        elif self.name == "Fire Elemental Skill":
            self.damage_type = DamageElementalType.PYRO
        else:
            raise AssertionError(f"Unknown name {self.name}")
        self.cost.elemental_dice_color = ELEMENT_TO_DIE_COLOR[
            DAMAGE_TYPE_TO_ELEMENT[self.damage_type]
        ]

    def get_actions(self, match: Match) -> List[Actions]:
        if self.damage_type == DamageElementalType.CRYO:
            status_name = "Sheer Cold"
        elif self.damage_type == DamageElementalType.PYRO:
            status_name = "Blazing Heat"
        else:
            raise AssertionError(f"Unknown damage type {self.damage_type}")
        return super().get_actions(
            match, [self.create_opposite_character_status(match, status_name)]
        )


class CarmineChrysalis(SignoraSkillValidCheck, ElementalBurstBase):
    name: Literal["Carmine Chrysalis"] = "Carmine Chrysalis"
    damage: int = 4
    damage_type: DamageElementalType = DamageElementalType.CRYO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.CRYO, elemental_dice_number=3, charge=2
    )

    def get_actions(self, match: Match) -> List[Actions]:
        """
        Damage 4, heal 2, and remove status.
        """
        damage = self.attack_opposite_active(match, self.damage, self.damage_type)
        heal = self.attack_self(match, -2)
        damage.damage_value_list += heal.damage_value_list
        status_list = (
            match.player_tables[self.position.player_idx]
            .characters[self.position.character_idx]
            .status
        )
        target_status = None
        for s in status_list:
            if s.name == "Ice-Sealed Crimson Witch of Embers":  # pragma: no branch
                target_status = s
                break
        else:
            raise AssertionError("Cannot find status")
        return [
            self.charge_self(-2),
            damage,
            RemoveObjectAction(object_position=target_status.position),
        ]


class SignoraPyroBurst(SignoraSkillValidCheck, ElementalBurstBase):
    name: Literal["Pyro Burst"] = "Pyro Burst"
    damage: int = 6
    damage_type: DamageElementalType = DamageElementalType.PYRO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.PYRO, elemental_dice_number=3, charge=2
    )


class MightOfDelusion(CreateStatusPassiveSkill):
    name: Literal["Might of Delusion"] = "Might of Delusion"
    status_name: Literal[
        "Ice-Sealed Crimson Witch of Embers"
    ] = "Ice-Sealed Crimson Witch of Embers"
    regenerate_when_revive: bool = False


class PainForPain_4_3(TalentBase):
    name: Literal["Pain for Pain"]
    version: Literal["4.3"] = "4.3"
    character_name: Literal["Signora"] = "Signora"
    cost_label: int = (
        CostLabels.CARD.value | CostLabels.TALENT.value | CostLabels.EQUIPMENT.value
    )
    cost: Cost = Cost(same_dice_number=3)

    def equip(self, match: Any) -> List[CreateDiceAction | CreateObjectAction]:
        """
        When equip, generate 3 element dice and create status.
        """
        character = match.player_tables[self.position.player_idx].characters[
            self.position.character_idx
        ]
        assert character.name == self.character_name
        return [
            CreateDiceAction(
                player_idx=self.position.player_idx,
                number=3,
                color=ELEMENT_TO_DIE_COLOR[character.element],
            ),
            CreateObjectAction(
                object_name=self.name,
                object_position=self.position.set_area(
                    ObjectPositionType.CHARACTER_STATUS
                ),
                object_arguments={},
            ),
        ]

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Match
    ) -> List[CreateObjectAction]:
        """
        When round prepare and equipped, create status.
        """
        if self.position.area != ObjectPositionType.CHARACTER:
            # not equipped, do nothing
            return []
        return [
            CreateObjectAction(
                object_name=self.name,
                object_position=self.position.set_area(
                    ObjectPositionType.CHARACTER_STATUS
                ),
                object_arguments={},
            )
        ]


class Signora_4_3(CharacterBase):
    name: Literal["Signora"]
    version: Literal["4.3"] = "4.3"
    desc: Literal["", "pyro"] = ""
    element: ElementType = ElementType.CRYO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        SignoraNormalAttack
        | SignoraElementalSkill
        | CarmineChrysalis
        | SignoraPyroBurst
        | MightOfDelusion
    ] = []
    faction: List[FactionType] = [FactionType.FATUI]
    weapon_type: WeaponType = WeaponType.OTHER

    def _init_skills(self) -> None:
        self.skills = [
            SignoraNormalAttack(name="Frostblade Hailstorm"),
            SignoraNormalAttack(name="Fire Normal Attack"),
            SignoraElementalSkill(name="Biting Shards"),
            SignoraElementalSkill(name="Fire Elemental Skill"),
            CarmineChrysalis(),
            SignoraPyroBurst(),
            MightOfDelusion(),
        ]

    def event_handler_REMOVE_OBJECT(
        self, event: RemoveObjectEventArguments, match: Match
    ) -> List[Actions]:
        """
        When Ice-Sealed Crimson Witch of Embers is removed, transform self
        element and desc.
        """
        status = (
            match.player_tables[self.position.player_idx]
            .characters[self.position.character_idx]
            .status
        )
        for s in status:
            if (  # pragma: no branch
                s.name == "Ice-Sealed Crimson Witch of Embers"
            ):
                # status still valid, do nothing
                return []
        # status not found, transform
        self.element = ElementType.PYRO
        self.desc = "pyro"
        return []


desc: Dict[str, DescDictType] = {
    "CHARACTER/Signora": {
        "names": {"en-US": "Signora", "zh-CN": "「女士」"},
        "descs": {"4.3": {"en-US": "", "zh-CN": ""}},
        "image_path": "cardface/Char_Monster_LaSignora.png",  # noqa: E501
        "id": 2102,
    },
    "CHARACTER/Signora_pyro": {
        "names": {"en-US": "Signora·Burned", "zh-CN": "「女士」·燃起来了"},
        "descs": {"4.3": {"en-US": "", "zh-CN": ""}},
        "image_path": "cardface/Char_Monster_LaSignoraHarbinger.png",  # noqa: E501
    },
    "SKILL_Signora_NORMAL_ATTACK/Frostblade Hailstorm": {
        "names": {"en-US": "Frostblade Hailstorm", "zh-CN": "霜锋霰舞"},
        "descs": {
            "4.3": {"en-US": "Deals 1 Cryo DMG.", "zh-CN": "造成1点冰元素伤害。"}
        },
    },
    "SKILL_Signora_NORMAL_ATTACK/Fire Normal Attack": {
        "names": {"en-US": "Crimson Lotus Moth", "zh-CN": "红莲之蛾"},
        "descs": {
            "4.3": {"en-US": "Deals 1 Pyro DMG.", "zh-CN": "造成1点火元素伤害。"}
        },
    },
    "SKILL_Signora_ELEMENTAL_SKILL/Biting Shards": {
        "names": {"en-US": "Biting Shards", "zh-CN": "凛冽之刺"},
        "descs": {
            "4.3": {
                "en-US": "Deals 2 Cryo DMG. The target character receives Sheer Cold.",  # noqa: E501
                "zh-CN": "造成2点冰元素伤害，目标角色附属严寒。",
            }
        },
    },
    "SKILL_Signora_ELEMENTAL_SKILL/Fire Elemental Skill": {
        "names": {"en-US": "Decimating Lash", "zh-CN": "烬灭之鞭"},
        "descs": {
            "4.3": {
                "en-US": "Deals 2 Pyro DMG. The target character receives Blazing Heat.",  # noqa: E501
                "zh-CN": "造成2点火元素伤害，目标角色附属炽热。",
            }
        },
    },
    "CHARACTER_STATUS/Sheer Cold": {
        "names": {"en-US": "Sheer Cold", "zh-CN": "严寒"},
        "descs": {
            "4.3": {
                "en-US": "End Phase: Deals 1 Cryo DMG to attached character.\nUsage(s): 1\n\nWhen attached character has Blazing Heat attached, remove this effect.",  # noqa: E501
                "zh-CN": "结束阶段：对附属角色造成1点冰元素伤害。\n可用次数：1\n\n所附属角色被附属炽热时，移除此效果。",  # noqa: E501
            }
        },
    },
    "CHARACTER_STATUS/Blazing Heat": {
        "names": {"en-US": "Blazing Heat", "zh-CN": "炽热"},
        "descs": {
            "4.3": {
                "en-US": "End Phase: Deals 1 Pyro DMG to attached character.\nUsage(s): 1\n\nWhen attached character has Sheer Cold attached, remove this effect.",  # noqa: E501
                "zh-CN": "结束阶段：对附属角色造成1点火元素伤害。\n可用次数：1\n\n所附属角色被附属严寒时，移除此效果。",  # noqa: E501
            }
        },
    },
    "SKILL_Signora_ELEMENTAL_BURST/Carmine Chrysalis": {
        "names": {"en-US": "Carmine Chrysalis", "zh-CN": "红莲冰茧"},
        "descs": {
            "4.3": {
                "en-US": "Deals 4 Cryo DMG, heals this character for 2 HP. Removes Ice-Sealed Crimson Witch of Embers. This character permanently switches to Crimson Witch of Embers state.",  # noqa: E501
                "zh-CN": "造成4点冰元素伤害，治疗本角色2点。移除冰封的炽炎魔女，本角色永久转换为「焚尽的炽炎魔女」形态。",  # noqa: E501
            }
        },
    },
    "SKILL_Signora_ELEMENTAL_BURST/Pyro Burst": {
        "names": {"en-US": "Whirling Blaze", "zh-CN": "燃焰旋织"},
        "descs": {
            "4.3": {"en-US": "Deals 6 Pyro DMG.", "zh-CN": "造成6点火元素伤害。"}
        },
    },
    "CHARACTER_STATUS/Ice-Sealed Crimson Witch of Embers": {
        "names": {
            "en-US": "Ice-Sealed Crimson Witch of Embers",
            "zh-CN": "冰封的炽炎魔女",
        },
        "descs": {
            "4.3": {
                "en-US": "When the Action Phase starts: If the attached character has no more than 4 HP, remove this effect.\nWhen attached character is defeated: Remove this effect, ensure that character $[K54|s1] and heal them to 1 HP.\n\nWhen this effect is removed: Attached character will transform into the Crimson Witch of Embers state.",  # noqa: E501
                "zh-CN": "行动阶段开始时：如果所附属角色生命值不多于4，则移除此效果。\n所附属角色被击倒时：移除此效果，使角色免于被击倒，并治疗该角色到1点生命值。\n\n此效果被移除时：所附属角色转换为「焚尽的炽炎魔女」形态。",  # noqa: E501
            }
        },
        "image_path": "status/LaSignora_S.png",
    },
    "SKILL_Signora_PASSIVE/Might of Delusion": {
        "names": {"en-US": "Might of Delusion", "zh-CN": "邪眼之威"},
        "descs": {
            "4.3": {
                "en-US": "(Passive) When the battle begins, this character gains Ice-Sealed Crimson Witch of Embers.",  # noqa: E501
                "zh-CN": "【被动】战斗开始时，初始附属冰封的炽炎魔女。",
            }
        },
    },
    "CHARACTER_STATUS/Pain for Pain": {
        "names": {"en-US": "Pain for Pain", "zh-CN": "苦痛奉还"},
        "descs": {
            "4.3": {
                "en-US": "When this character takes at least 3 DMG: Negates 1 DMG, then apply Sheer Cold or Blazing Heat to the opposing active character based on Signora's state. (Once per Round)",  # noqa: E501
                "zh-CN": "角色受到至少为3点的伤害时：抵消1点伤害，然后根据「女士」的形态对敌方出战角色附属严寒或炽热。（每回合1次）",  # noqa: E501
            }
        },
    },
    "TALENT_Signora/Pain for Pain": {
        "names": {"en-US": "Pain for Pain", "zh-CN": "苦痛奉还"},
        "descs": {
            "4.3": {
                "en-US": "Can only be played if your active character is Signora: When entering play, generate 3 Elemental Dice of the same Type as Signora.\nWhen this character takes at least 3 DMG: Negates 1 DMG, then apply Sheer Cold or Blazing Heat to the opposing active character based on Signora's state. (Once per Round)\n(Your deck must contain Signora to add this card to your deck)",  # noqa: E501
                "zh-CN": "我方出战角色为「女士」时，才能打出：入场时，生成3个「女士」当前元素类型的元素骰。\n角色受到至少为3点的伤害时：抵消1点伤害，然后根据「女士」的形态对敌方出战角色附属严寒或炽热。（每回合1次）\n（牌组中包含「女士」，才能加入牌组）",  # noqa: E501
            }
        },
        "image_path": "cardface/Modify_Talent_LaSignora.png",  # noqa: E501
        "id": 221021,
    },
}


register_class(
    IceSealedCrimsonWitchOfEmbers_4_3
    | SheerCold_4_3
    | Signora_4_3
    | PainForPain_4_3
    | PainForPainStatus_4_3,
    desc,
)



from typing import Any, List, Literal

from ...event import CreateObjectEventArguments, RemoveObjectEventArguments

from ....utils.class_registry import register_class

from ...action import Actions

from ...struct import Cost

from ...consts import (
    DamageElementalType,
    DieColor,
    ElementType,
    FactionType,
    WeaponType,
)
from ..character_base import (
    ElementalBurstBase,
    ElementalSkillBase,
    PhysicalNormalAttackBase,
    CharacterBase,
    SkillTalent,
)


# Skills


class BaneOfAllEvil(ElementalBurstBase):
    name: Literal["Bane of All Evil"] = "Bane of All Evil"
    damage: int = 4
    damage_type: DamageElementalType = DamageElementalType.ANEMO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.ANEMO, elemental_dice_number=3, charge=2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        If has talent, add skill cost decrease by 2
        """
        args = {}
        if self.is_talent_equipped(match):
            args["skill_cost_decrease_usage"] = 2
        return super().get_actions(
            match,
            [
                self.create_character_status("Yaksha's Mask", args),
            ],
        )


# Talents


class ConquerorOfEvilGuardianYaksha_3_7(SkillTalent):
    name: Literal["Conqueror of Evil: Guardian Yaksha"]
    version: Literal["3.7"] = "3.7"
    character_name: Literal["Xiao"] = "Xiao"
    cost: Cost = Cost(
        elemental_dice_color=DieColor.ANEMO, elemental_dice_number=3, charge=2
    )
    skill: Literal["Bane of All Evil"] = "Bane of All Evil"


# character base


class Xiao_3_7(CharacterBase):
    name: Literal["Xiao"]
    desc: Literal["", "transform"] = ""
    version: Literal["3.7"] = "3.7"
    element: ElementType = ElementType.ANEMO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[PhysicalNormalAttackBase | ElementalSkillBase | BaneOfAllEvil] = []
    faction: List[FactionType] = [FactionType.LIYUE]
    weapon_type: WeaponType = WeaponType.POLEARM

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name="Whirlwind Thrust",
                cost=PhysicalNormalAttackBase.get_cost(self.element),
            ),
            ElementalSkillBase(
                name="Lemniscatic Wind Cycling",
                damage_type=DamageElementalType.ANEMO,
                cost=ElementalSkillBase.get_cost(self.element),
            ),
            BaneOfAllEvil(),
        ]

    def _update_desc(self, match):
        status = (
            match.player_tables[self.position.player_idx]
            .characters[self.position.character_idx]
            .status
        )
        for s in status:
            if s.name == "Yaksha's Mask":
                self.desc = "transform"
                return
        else:
            self.desc = ""

    def event_handler_REMOVE_OBJECT(
        self, event: RemoveObjectEventArguments, match: Any
    ) -> List[Actions]:
        self._update_desc(match)
        return []

    def event_handler_CREATE_OBJECT(
        self, event: CreateObjectEventArguments, match: Any
    ) -> List[Actions]:
        self._update_desc(match)
        return []


register_class(Xiao_3_7 | ConquerorOfEvilGuardianYaksha_3_7)



"""
Event cards that not belong to any other categories.
"""

from typing import Any, List, Literal, Tuple

from ....utils.class_registry import register_class

from ...consts import (
    DieColor,
    ObjectPositionType,
    ObjectType,
    PlayerActionLabels,
    SkillType,
)

from ...object_base import (
    CreateSystemEventHandlerObject,
    EventCardBase,
    MultiTargetEventCardBase,
)
from ...action import (
    ActionTypes,
    ChangeObjectUsageAction,
    ChargeAction,
    CreateDiceAction,
    CreateObjectAction,
    DrawCardAction,
    GenerateRerollDiceRequestAction,
    MoveObjectAction,
    RemoveObjectAction,
    SkillEndAction,
    SwitchCharacterAction,
    UseSkillAction,
)
from ...struct import Cost, MultipleObjectPosition, ObjectPosition


class TheBestestTravelCompanion_3_3(EventCardBase):
    name: Literal["The Bestest Travel Companion!"]
    version: Literal["3.3"] = "3.3"
    cost: Cost = Cost(any_dice_number=2)

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        # no targets
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[CreateDiceAction]:
        """
        Act the card. Convert the Elemental Dice spent to Omni Element x2.
        """
        assert target is None  # no targets
        return [
            CreateDiceAction(
                player_idx=self.position.player_idx,
                color=DieColor.OMNI,
                number=2,
            )
        ]


class ChangingShifts_3_3(EventCardBase):
    name: Literal["Changing Shifts"]
    version: Literal["3.3"] = "3.3"
    cost: Cost = Cost()

    def is_valid(self, match: Any) -> bool:
        """
        When only one character is alive, cannot use this card.
        """
        characters = match.player_tables[self.position.player_idx].characters
        counter = 0
        for character in characters:
            if character.is_alive:
                counter += 1
        return counter > 1

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        # no targets
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[CreateObjectAction]:
        """
        Act the card. Create team status.
        """
        assert target is None  # no targets
        return [
            CreateObjectAction(
                object_name=self.name,
                object_position=ObjectPosition(
                    player_idx=self.position.player_idx,
                    area=ObjectPositionType.TEAM_STATUS,
                    id=-1,
                ),
                object_arguments={},
            )
        ]


class TossUp_3_3(EventCardBase):
    name: Literal["Toss-Up"]
    version: Literal["3.3"] = "3.3"
    cost: Cost = Cost()

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        # no targets
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[GenerateRerollDiceRequestAction]:
        assert target is None
        return [
            GenerateRerollDiceRequestAction(
                player_idx=self.position.player_idx,
                reroll_times=2,
            )
        ]


class Strategize_3_3(EventCardBase):
    name: Literal["Strategize"]
    version: Literal["3.3"] = "3.3"
    cost: Cost = Cost(same_dice_number=1)

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        # no targets
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[DrawCardAction]:
        """
        Act the card. Draw two cards.
        """
        assert target is None  # no targets
        return [
            DrawCardAction(
                player_idx=self.position.player_idx,
                number=2,
                draw_if_filtered_not_enough=True,
            )
        ]


class IHaventLostYet_4_0(CreateSystemEventHandlerObject, EventCardBase):
    name: Literal["I Haven't Lost Yet!"]
    version: Literal["4.0"] = "4.0"
    cost: Cost = Cost()

    activated: bool = False

    handler_name: str = "I Haven't Lost Yet!"

    def is_valid(self, match: Any) -> bool:
        team_status = match.player_tables[self.position.player_idx].team_status
        for status in team_status:
            if status.name == self.name:
                # activated in this round
                return False
        handler = self._get_event_handler(match)
        return handler.defeated[self.position.player_idx]

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        # no targets
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[ChargeAction | CreateDiceAction | CreateObjectAction]:
        """
        1 omni die, 1 energy, and generate used status.
        """
        assert target is None
        return [
            CreateDiceAction(
                player_idx=self.position.player_idx,
                color=DieColor.OMNI,
                number=1,
            ),
            ChargeAction(
                player_idx=self.position.player_idx,
                character_idx=match.player_tables[
                    self.position.player_idx
                ].active_character_idx,
                charge=1,
            ),
            CreateObjectAction(
                object_name=self.name,
                object_position=ObjectPosition(
                    player_idx=self.position.player_idx,
                    area=ObjectPositionType.TEAM_STATUS,
                    id=-1,
                ),
                object_arguments={},
            ),
        ]


class IHaventLostYet_3_3(IHaventLostYet_4_0):
    name: Literal["I Haven't Lost Yet!"]
    version: Literal["3.3"]
    cost: Cost = Cost()

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[ChargeAction | CreateDiceAction | CreateObjectAction]:
        ret = super().get_actions(target, match)
        assert ret[-1].type == ActionTypes.CREATE_OBJECT
        # old version not adding team status
        ret = ret[:-1]
        return ret


class LeaveItToMe_3_3(EventCardBase):
    name: Literal["Leave It to Me!"]
    version: Literal["3.3"] = "3.3"
    cost: Cost = Cost()

    def is_valid(self, match: Any) -> bool:
        """
        When only one character is alive, cannot use this card.
        """
        characters = match.player_tables[self.position.player_idx].characters
        counter = 0
        for character in characters:
            if character.is_alive:
                counter += 1
        return counter > 1

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        # no targets
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[CreateObjectAction]:
        """
        Act the card. Create team status.
        """
        assert target is None  # no targets
        return [
            CreateObjectAction(
                object_name=self.name,
                object_position=ObjectPosition(
                    player_idx=self.position.player_idx,
                    area=ObjectPositionType.TEAM_STATUS,
                    id=-1,
                ),
                object_arguments={},
            )
        ]


class WhenTheCraneReturned_3_3(EventCardBase):
    name: Literal["When the Crane Returned"]
    version: Literal["3.3"] = "3.3"
    cost: Cost = Cost(same_dice_number=1)

    def is_valid(self, match: Any) -> bool:
        """
        When only one character is alive, cannot use this card.
        """
        characters = match.player_tables[self.position.player_idx].characters
        counter = 0
        for character in characters:
            if character.is_alive:
                counter += 1
        return counter > 1

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        # no targets
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[CreateObjectAction]:
        """
        Act the card. Create team status.
        """
        assert target is None  # no targets
        return [
            CreateObjectAction(
                object_name=self.name,
                object_position=ObjectPosition(
                    player_idx=self.position.player_idx,
                    area=ObjectPositionType.TEAM_STATUS,
                    id=-1,
                ),
                object_arguments={},
            )
        ]


class Starsigns_3_3(EventCardBase):
    name: Literal["Starsigns"]
    version: Literal["3.3"] = "3.3"
    cost: Cost = Cost(any_dice_number=2)

    def is_valid(self, match: Any) -> bool:
        """
        can use if charge not full
        """
        active_character = match.player_tables[
            self.position.player_idx
        ].get_active_character()
        return active_character.charge < active_character.max_charge

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        # active character
        return [
            match.player_tables[self.position.player_idx]
            .get_active_character()
            .position
        ]

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[ChargeAction]:
        assert target is not None
        return [
            ChargeAction(
                player_idx=target.player_idx,
                character_idx=target.character_idx,
                charge=1,
            )
        ]


class ClaxsArts_3_3(EventCardBase):
    name: Literal["Calx's Arts"]
    version: Literal["3.3"] = "3.3"
    cost: Cost = Cost(same_dice_number=1)

    def _get_charge_source_and_targets(self, match: Any) -> Tuple[int, List[int]]:
        """
        Get charge source and targets.
        """
        table = match.player_tables[self.position.player_idx]
        source = table.active_character_idx
        assert source >= 0 and source < len(table.characters)
        targets = []
        for idx, character in enumerate(table.characters):
            if idx != source and character.charge > 0:
                targets.append(idx)
        if len(targets) > 2:
            targets = targets[:2]
        return source, targets

    def is_valid(self, match: Any) -> bool:
        source, targets = self._get_charge_source_and_targets(match)
        return len(targets) > 0

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        # no targets
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[ChargeAction]:
        assert target is None
        source, targets = self._get_charge_source_and_targets(match)
        assert len(targets) > 0
        ret: List[ChargeAction] = []
        ret.append(
            ChargeAction(
                player_idx=self.position.player_idx,
                character_idx=source,
                charge=len(targets),
            )
        )
        for t in targets:
            ret.append(
                ChargeAction(
                    player_idx=self.position.player_idx,
                    character_idx=t,
                    charge=-1,
                )
            )
        return ret


class MasterOfWeaponry_4_1(MultiTargetEventCardBase):
    name: Literal["Master of Weaponry"]
    version: Literal["4.1"] = "4.1"
    cost: Cost = Cost()
    move_type: Literal[ObjectType.WEAPON, ObjectType.ARTIFACT] = ObjectType.WEAPON
    reset_usage: bool = True

    def _move_target_id(self, character: Any) -> int | None:
        """
        If has target, return target id. otherwise return None
        """
        if self.move_type == ObjectType.WEAPON:
            return None if character.weapon is None else character.weapon.id
        elif self.move_type == ObjectType.ARTIFACT:
            return None if character.artifact is None else character.artifact.id
        else:
            raise AssertionError("Unknown move type")

    def get_targets(self, match: Any) -> List[MultipleObjectPosition]:
        characters = match.player_tables[self.position.player_idx].characters
        ret: List[MultipleObjectPosition] = []
        for character in characters:
            target_id = self._move_target_id(character)
            if target_id is not None:
                for target in characters:
                    if target.id != character.id:
                        if (
                            self.move_type == ObjectType.WEAPON
                            and character.weapon_type != target.weapon_type
                        ):
                            # weapon type not right
                            continue
                        ret.append(
                            MultipleObjectPosition(
                                positions=[
                                    character.position.set_id(target_id),
                                    target.position.set_id(target_id),
                                ]
                            )
                        )
        return ret

    def is_valid(self, match: Any) -> bool:
        return len(self.get_targets(match)) > 0

    def get_actions(
        self, target: MultipleObjectPosition | None, match: Any
    ) -> List[RemoveObjectAction | MoveObjectAction]:
        assert target is not None
        assert len(target.positions) == 2
        ret: List[RemoveObjectAction | MoveObjectAction] = []
        target_character = match.player_tables[
            target.positions[1].player_idx
        ].characters[target.positions[1].character_idx]
        target_equip: Any = None
        if self.move_type == ObjectType.WEAPON:
            target_equip = target_character.weapon
        elif self.move_type == ObjectType.ARTIFACT:
            target_equip = target_character.artifact
        else:
            raise AssertionError("Unknown move type")
        if target_equip is not None:
            ret.append(RemoveObjectAction(object_position=target_equip.position))
        return ret + [
            MoveObjectAction(
                object_position=target.positions[0],
                target_position=target.positions[1],
                reset_usage=self.reset_usage,
            )
        ]


class BlessingOfTheDivineRelicsInstallation_4_1(MasterOfWeaponry_4_1):
    name: Literal["Blessing of the Divine Relic's Installation"]
    version: Literal["4.1"] = "4.1"
    cost: Cost = Cost()
    move_type: Literal[ObjectType.WEAPON, ObjectType.ARTIFACT] = ObjectType.ARTIFACT
    reset_usage: bool = True


class MasterOfWeaponry_3_3(MasterOfWeaponry_4_1):
    version: Literal["3.3"]
    reset_usage: bool = False


class BlessingOfTheDivineRelicsInstallation_3_3(
    BlessingOfTheDivineRelicsInstallation_4_1
):
    version: Literal["3.3"]
    reset_usage: bool = False


class QuickKnit_3_3(EventCardBase):
    name: Literal["Quick Knit"]
    version: Literal["3.3"] = "3.3"
    cost: Cost = Cost(same_dice_number=1)

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        # all current summons
        return [
            x.position for x in match.player_tables[self.position.player_idx].summons
        ]

    def is_valid(self, match: Any) -> bool:
        # should have summon
        return len(self.get_targets(match)) > 0

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[ChangeObjectUsageAction]:
        """
        Add 1 usage of selected summon
        """
        assert target is not None
        return [ChangeObjectUsageAction(object_position=target, change_usage=1)]


class SendOff_3_7(EventCardBase):
    name: Literal["Send Off"]
    version: Literal["3.7"] = "3.7"
    cost: Cost = Cost(same_dice_number=2)

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        """
        Opposing summons
        """
        summons = match.player_tables[1 - self.position.player_idx].summons
        return [x.position for x in summons]

    def is_valid(self, match: Any) -> bool:
        return len(self.get_targets(match)) > 0

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[ChangeObjectUsageAction]:
        """
        Act the card. Create team status.
        """
        assert target is not None
        return [ChangeObjectUsageAction(object_position=target, change_usage=-2)]


class SendOff_3_3(SendOff_3_7):
    name: Literal["Send Off"]
    version: Literal["3.3"]
    cost: Cost = Cost(any_dice_number=2)

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[RemoveObjectAction]:
        """
        Act the card. Create team status.
        """
        assert target is not None
        return [
            RemoveObjectAction(
                object_position=target,
            )
        ]


class GuardiansOath_3_3(EventCardBase):
    name: Literal["Guardian's Oath"]
    version: Literal["3.3"] = "3.3"
    cost: Cost = Cost(same_dice_number=4)

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        # no targets
        return []

    def is_valid(self, match: Any) -> bool:
        return (
            len(match.player_tables[0].summons) > 0
            or len(match.player_tables[1].summons) > 0
        )

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[RemoveObjectAction]:
        """
        Act the card. Destroy all summons.
        """
        assert target is None
        ret: List[RemoveObjectAction] = []
        for table in match.player_tables:
            for summon in table.summons:
                ret.append(RemoveObjectAction(object_position=summon.position))
        return ret


class PlungingStrike_3_7(EventCardBase):
    name: Literal["Plunging Strike"]
    version: Literal["3.7"] = "3.7"
    cost: Cost = Cost(same_dice_number=3)

    def get_action_type(self, match: Any) -> Tuple[int, bool]:
        return (
            PlayerActionLabels.SWITCH.value
            | PlayerActionLabels.CARD.value
            | PlayerActionLabels.SKILL.value,
            True,
        )

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        """ """
        pos: List[ObjectPosition] = []
        table = match.player_tables[self.position.player_idx]
        for cid, character in enumerate(table.characters):
            if character.is_alive and cid != table.active_character_idx:
                pos.append(character.position)
        return pos

    def is_valid(self, match: Any) -> bool:
        return len(self.get_targets(match)) > 0

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[SwitchCharacterAction | UseSkillAction | SkillEndAction]:
        """
        Switch to target, and use target's normal attack
        """
        assert target is not None
        skills = (
            match.player_tables[target.player_idx]
            .characters[target.character_idx]
            .skills
        )
        normal_idx: int = -1
        for idx, skill in enumerate(skills):
            if skill.skill_type == SkillType.NORMAL_ATTACK:
                assert normal_idx == -1, "Multiple normal attack skill"
                normal_idx = idx
        return [
            SwitchCharacterAction(
                player_idx=target.player_idx, character_idx=target.character_idx
            ),
            UseSkillAction(
                skill_position=skills[normal_idx].position,
            ),
            SkillEndAction(
                position=skills[normal_idx].position,
                target_position=match.player_tables[1 - target.player_idx]
                .get_active_character()
                .position,
                skill_type=SkillType.NORMAL_ATTACK,
            ),
        ]


class HeavyStrike_3_7(EventCardBase):
    name: Literal["Heavy Strike"]
    version: Literal["3.7"] = "3.7"
    cost: Cost = Cost(same_dice_number=1)

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        # active character
        return [
            match.player_tables[self.position.player_idx]
            .get_active_character()
            .position
        ]

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[CreateObjectAction]:
        """
        Act the card. Create character status.
        """
        assert target is not None
        return [
            CreateObjectAction(
                object_name=self.name,
                object_position=target.set_area(ObjectPositionType.CHARACTER_STATUS),
                object_arguments={},
            )
        ]


class TheLegendOfVennessa_3_7(EventCardBase):
    name: Literal["The Legend of Vennessa"]
    version: Literal["3.7"] = "3.7"
    cost: Cost = Cost(same_dice_number=3)

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        # no targets
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[CreateDiceAction]:
        """
        Act the card. Create 4 basic Elemental Dice of different types.
        """
        assert target is None
        return [
            CreateDiceAction(
                player_idx=self.position.player_idx, number=4, different=True
            ),
        ]


class FriendshipEternal_3_7(EventCardBase):
    name: Literal["Friendship Eternal"]
    version: Literal["3.7"] = "3.7"
    cost: Cost = Cost(same_dice_number=2)

    def is_valid(self, match: Any) -> bool:
        # if both player have 4 or more cards, cannot use this card.
        for pidx, table in enumerate(match.player_tables):
            if len(table.hands) < 4:
                return True
            if len(table.hands) == 4 and pidx == self.position.player_idx:
                # for self, 4 card can use to draw one card
                return True
        return False

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        # no targets
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[DrawCardAction]:
        """
        Act the card. Both player draw card until card number reach 4.
        """
        assert target is None  # no targets
        ret: List[DrawCardAction] = []
        for pidx, table in enumerate(match.player_tables):
            target_number = 4
            if len(table.hands) >= target_number:
                # already have enough cards, do nothing
                continue
            ret.append(
                DrawCardAction(
                    player_idx=pidx,
                    number=target_number - len(table.hands),
                    draw_if_filtered_not_enough=True,
                )
            )
        return ret


class RhythmOfTheGreatDream_3_8(EventCardBase):
    name: Literal["Rhythm of the Great Dream"]
    version: Literal["3.8"] = "3.8"
    cost: Cost = Cost()

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        # no targets
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[CreateObjectAction]:
        """
        Act the card. Create team status.
        """
        assert target is None
        return [
            CreateObjectAction(
                object_name=self.name,
                object_position=ObjectPosition(
                    player_idx=self.position.player_idx,
                    area=ObjectPositionType.TEAM_STATUS,
                    id=-1,
                ),
                object_arguments={},
            )
        ]


class WhereIstheUnseenRazor_4_0(EventCardBase):
    name: Literal["Where Is the Unseen Razor?"]
    version: Literal["4.0"] = "4.0"
    cost: Cost = Cost()

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        """
        Self characters that have weapon.
        """
        characters = match.player_tables[self.position.player_idx].characters
        res: List[ObjectPosition] = []
        for character in characters:
            if character.weapon is not None:
                res.append(character.position)
        return res

    def is_valid(self, match: Any) -> bool:
        return len(self.get_targets(match)) > 0

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[MoveObjectAction | CreateObjectAction]:
        assert target is not None
        character = match.get_object(target)
        assert character is not None
        assert character.weapon is not None
        target_position = character.weapon.position.set_area(ObjectPositionType.HAND)
        return [
            MoveObjectAction(
                object_position=character.weapon.position,
                target_position=target_position,
            ),
            CreateObjectAction(
                object_name=self.name,
                object_position=ObjectPosition(
                    player_idx=self.position.player_idx,
                    area=ObjectPositionType.TEAM_STATUS,
                    id=-1,
                ),
                object_arguments={},
            ),
        ]


class Pankration_4_1(EventCardBase):
    name: Literal["Pankration!"]
    version: Literal["4.1"] = "4.1"
    cost: Cost = Cost()

    def is_valid(self, match: Any) -> bool:
        return (
            len(match.player_tables[self.position.player_idx].dice.colors) >= 8
            and not match.player_tables[1 - self.position.player_idx].has_round_ended
        )

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[CreateObjectAction]:
        """
        create status in our side
        """
        assert target is None
        return [
            CreateObjectAction(
                object_position=ObjectPosition(
                    player_idx=self.position.player_idx,
                    area=ObjectPositionType.TEAM_STATUS,
                    id=-1,
                ),
                object_name=self.name,
                object_arguments={},
            )
        ]


class Lyresong_4_2(EventCardBase):
    name: Literal["Lyresong"]
    version: Literal["4.2"] = "4.2"
    cost: Cost = Cost()

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        """
        Self characters that have artifact.
        """
        characters = match.player_tables[self.position.player_idx].characters
        res: List[ObjectPosition] = []
        for character in characters:
            if character.artifact is not None:
                res.append(character.position)
        return res

    def is_valid(self, match: Any) -> bool:
        return len(self.get_targets(match)) > 0

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[MoveObjectAction | CreateObjectAction]:
        assert target is not None
        character = match.get_object(target)
        assert character is not None
        assert character.artifact is not None
        target_position = character.artifact.position.set_area(ObjectPositionType.HAND)
        return [
            MoveObjectAction(
                object_position=character.artifact.position,
                target_position=target_position,
            ),
            CreateObjectAction(
                object_name=self.name,
                object_position=ObjectPosition(
                    player_idx=self.position.player_idx,
                    area=ObjectPositionType.TEAM_STATUS,
                    id=-1,
                ),
                object_arguments={},
            ),
        ]


register_class(
    TheBestestTravelCompanion_3_3
    | ChangingShifts_3_3
    | TossUp_3_3
    | Strategize_3_3
    | IHaventLostYet_4_0
    | LeaveItToMe_3_3
    | WhenTheCraneReturned_3_3
    | Starsigns_3_3
    | ClaxsArts_3_3
    | MasterOfWeaponry_4_1
    | BlessingOfTheDivineRelicsInstallation_4_1
    | QuickKnit_3_3
    | SendOff_3_7
    | GuardiansOath_3_3
    | PlungingStrike_3_7
    | HeavyStrike_3_7
    | TheLegendOfVennessa_3_7
    | FriendshipEternal_3_7
    | RhythmOfTheGreatDream_3_8
    | WhereIstheUnseenRazor_4_0
    | Pankration_4_1
    | Lyresong_4_2
    | IHaventLostYet_3_3
    | MasterOfWeaponry_3_3
    | BlessingOfTheDivineRelicsInstallation_3_3
    | SendOff_3_3
)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...action import Actions
from ...struct import Cost

from ...consts import (
    DamageElementalType,
    DieColor,
    ElementType,
    FactionType,
    WeaponType,
)
from ..character_base import (
    ElementalBurstBase,
    ElementalSkillBase,
    CharacterBase,
    PhysicalNormalAttackBase,
    SkillTalent,
)


class Breastplate(ElementalSkillBase):
    name: Literal["Breastplate"] = "Breastplate"
    damage: int = 1
    damage_type: DamageElementalType = DamageElementalType.GEO
    cost: Cost = Cost(elemental_dice_number=3, elemental_dice_color=DieColor.GEO)

    def get_actions(self, match: Any) -> List[Actions]:
        ret = super().get_actions(match, [self.create_team_status("Full Plate")])
        return ret


class SweepingTime(ElementalBurstBase):
    name: Literal["Sweeping Time"] = "Sweeping Time"
    damage: int = 4
    damage_type: DamageElementalType = DamageElementalType.GEO
    cost: Cost = Cost(
        elemental_dice_number=4, elemental_dice_color=DieColor.GEO, charge=2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        ret = super().get_actions(
            match, [self.create_character_status("Sweeping Time")]
        )
        return ret


class IGotYourBack_3_3(SkillTalent):
    name: Literal["I Got Your Back"]
    version: Literal["3.3"] = "3.3"
    character_name: Literal["Noelle"] = "Noelle"
    cost: Cost = Cost(elemental_dice_number=3, elemental_dice_color=DieColor.GEO)
    skill: Literal["Breastplate"] = "Breastplate"


class Noelle_3_3(CharacterBase):
    name: Literal["Noelle"]
    version: Literal["3.3"] = "3.3"
    element: ElementType = ElementType.GEO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[PhysicalNormalAttackBase | Breastplate | SweepingTime] = []
    faction: List[FactionType] = [FactionType.MONDSTADT]
    weapon_type: WeaponType = WeaponType.CLAYMORE

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name="Favonius Bladework - Maid",
                cost=PhysicalNormalAttackBase.get_cost(
                    ElementType.GEO,
                ),
            ),
            Breastplate(),
            SweepingTime(),
        ]


register_class(Noelle_3_3 | IGotYourBack_3_3)



from lpsim.agents.interaction_agent import InteractionAgent
from lpsim.server.match import Match, MatchState
from lpsim.server.deck import Deck
from tests.utils_for_test import (
    check_hp,
    get_random_state,
    get_test_id_from_command,
    make_respond,
    set_16_omni,
)


def test_maguu_kenki():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "skill 1 0 1 2",
            "sw_char 1 0",
            "skill 0 0 1 2",
            "skill 0 0 1 2",
            "skill 3 0 1 2",
            "sw_char 0 0",
            "TEST 1 10 4 10 10 4 10",
            "sw_char 1 0",
            "end",
            "sw_char 0 0",
            "skill 1 0 1 2",
            "TEST 1 3 2 9 9 4 10",
            "TEST 2 p1 summon usage 2 2",
            "sw_char 2 0",
            "sw_char 1 0",
            "choose 2",
            "sw_char 0 0",
            "card 0 0 0 1 2",
            "end",
            "sw_char 0 0",
            "sw_char 2 0",
            "sw_char 0 0",
            "TEST 3 p1 active 2",
            "sw_char 2 0",
            "skill 0 0 1 2",
            "choose 0",
            "skill 1 0 1 2",
            "skill 2 0 1 2",
            "skill 0 0 1 2",
            "end",
        ],
        [
            "sw_card",
            "choose 0",
            "skill 2 0 1 2",
            "sw_char 1 0",
            "skill 0 0 1 2",
            "skill 0 0 1 2",
            "skill 3 0 1 2",
            "sw_char 0 0",
            "end",
            "TEST 1 10 3 10 9 4 10",
            "skill 1 0 1 2",
            "skill 2 0 1 2",
            "skill 3 0 1 2",
            "sw_char 2 0",
            "skill 1 0 1 2",
            "end",
            "sw_char 0 0",
            "card 0 0 0 1 2",
            "TEST 3 p1 active 1",
            "sw_char 0 0",
            "skill 2 0 1 2",
            "sw_char 1 0",
            "skill 1 0 1 2",
            "end",
            "choose 0",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Maguu Kenki
        character:Nahida
        character:Fischl
        Transcendent Automaton*30
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                hps = cmd.strip().split(" ")[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            elif test_id == 2:
                summon = match.player_tables[1].summons
                assert len(summon) == 2
                for s in summon:
                    assert s.usage == 2
            elif test_id == 3:
                cmd = cmd.strip().split(" ")
                pidx = int(cmd[2][1])
                aidx = int(cmd[4])
                assert match.player_tables[pidx].active_character_idx == aidx
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_maguu_kenki_old():
    cmd_records = [
        [
            "sw_card",
            "choose 2",
            "card 0 0 0 1 2",
            "sw_char 2 9",
            "skill 2 0 1 2",
            "sw_char 2 0",
            "skill 0 0 1 2",
            "end",
            "choose 1",
            "TEST 1 7 8 0 9 10 5",
            "end",
            "choose 0",
            "TEST 1 6 0 0 9 8 5",
            "end",
            "TEST 1 2 0 0 9 8 5",
            "end",
        ],
        [
            "sw_card",
            "choose 2",
            "skill 2 0 1 2",
            "card 0 0 0 1 2",
            "sw_char 2 0",
            "skill 0 0 1 2",
            "skill 3 0 1 2",
            "end",
            "skill 0 0 1 2",
            "skill 0 0 1 2",
            "skill 0 0 1 2",
            "sw_char 1 0",
            "end",
            "sw_char 2 0",
            "skill 3 0 1 2",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Nahida
        character:Fischl
        Transcendent Automaton*30
        """
    )
    old = {"name": "Maguu Kenki", "version": "3.3"}
    deck_dict = deck.dict()
    deck_dict["characters"] += [old]
    deck = Deck(**deck_dict)
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                hps = cmd.strip().split(" ")[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


if __name__ == "__main__":
    test_maguu_kenki_old()
    test_maguu_kenki()



from lpsim.server.struct import ObjectPosition
from lpsim.agents.interaction_agent import InteractionAgent
from lpsim.server.match import Match, MatchState
from lpsim.server.deck import Deck
from tests.utils_for_test import (
    check_hp,
    get_random_state,
    get_test_id_from_command,
    make_respond,
    set_16_omni,
)


def test_woven_element():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "skill 0 0 1 2",
            "skill 0 0 1 2",
            "TEST 1 dendro pyro anemo hydro dendro " "cryo dendro electro dendro pyro",
            "card 0 0",
            "card 0 0",
            "card 0 0",
            "card 0 0",
            "card 0 0",
            "end",
        ],
        [
            "sw_card",
            "choose 0",
            "skill 0 0 1 2",
            "skill 0 0 1 2",
            "card 0 0",
            "card 0 0",
            "card 0 0",
            "card 0 0",
            "card 0 0",
            "TEST 2 color match",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Nahida*3
        Elemental Resonance: Woven Flames*4
        Elemental Resonance: Woven Ice*4
        Elemental Resonance: Woven Stone*4
        Elemental Resonance: Woven Thunder*4
        Elemental Resonance: Woven Waters*4
        Elemental Resonance: Woven Weeds*4
        Elemental Resonance: Woven Winds*4
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    record = []
    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                colors = [x.upper() for x in cmd.split(" ")[2:]]
                colors = [colors[:5], colors[5:]]
                for i in range(2):
                    for j in range(5):
                        assert match.player_tables[i].hands[j].element == colors[i][j]  # type: ignore
                record = colors
            elif test_id == 2:
                for table, rec in zip(match.player_tables, record):
                    m = {}
                    for r in rec:
                        m[r] = m.get(r, 0) + 1
                    for color in table.dice.colors:
                        if color != "OMNI":
                            m[color.name] -= 1
                    for v in m.values():
                        assert v == 0
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_enduring_rock():
    cmd_records = [
        [
            "sw_card",
            "choose 2",
            "skill 1 0 1 2",
            "card 2 0 0",
            "skill 0 0 1 2",
            "TEST 2 p1 5 full plate",
            "sw_char 0 0",
            "skill 0 0 1 2",
            "sw_char 2 0",
            "skill 2 0 1 2 3",
            "TEST 1 9 10 10 9 10 7",
            "TEST 3 p0 4 crystallize",
            "end",
            "card 3 0 0",
            "skill 0 0 1",
            "end",
            "end",
            "end",
            "card 3 0 0",
            "sw_char 1 0",
            "skill 1 0 1 2",
            "TEST 3 p0 1 elemental resonance: enduring rock",
            "end",
            "card 6 0 15",
            "skill 1 14 13 12",
            "sw_char 2 11",
            "skill 1 10 9 8",
            "TEST 7 p1 usage 4 1",
            "end",
        ],
        [
            "sw_card 1 2 3",
            "choose 2",
            "card 0 0 0",
            "skill 0 0 1 2",
            "TEST 4 p1 1 team status",
            "TEST 5 p0 1 full plate 1 geo",
            "skill 1 0 1 2",
            "sw_char 0 0",
            "skill 0 0 1 2",
            "end",
            "skill 0 0 1 2",
            "TEST 3 p0 6 crystallize",
            "card 0 0 0",
            "card 4 0 0",
            "sw_char 1 0",
            "card 0 0 0 1 2",
            "skill 0 0 1 2",
            "TEST 6 p1 1 rebellious shield",
            "end",
            "end",
            "card 5 0 0",
            "skill 1 0 1 2",
            "TEST 6 p1 5 rebellious shield",
            "end",
            "end",
            "card 5 0 15",
            "sw_char 2 14",
            "sw_char 1 13",
            "skill 0 12 11 10",
            "skill 2 9 8 7",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:PyroMobMage
        character:Arataki Itto
        character:Noelle
        Elemental Resonance: Enduring Rock*15
        The Bell*15
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                hps = [int(x) for x in cmd.split(" ")[2:]]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            elif test_id in [2, 3, 6]:
                cmd = cmd.split(" ")
                pid = int(cmd[2][1])
                usage = int(cmd[3])
                found = False
                for status in match.player_tables[pid].team_status:
                    if status.name.lower() == " ".join(cmd[4:]):
                        assert status.usage == usage
                        found = True
                assert found
            elif test_id == 4:
                assert len(match.player_tables[1].team_status) == 1
            elif test_id == 5:
                assert len(match.player_tables[0].team_status) == 2
                assert match.player_tables[0].team_status[0].usage == 1
                assert match.player_tables[0].team_status[1].usage == 1
            elif test_id == 7:
                cmd = cmd.split()
                pid = int(cmd[2][1])
                usage = [int(x) for x in cmd[4:]]
                status = match.player_tables[pid].team_status
                assert len(status) == len(usage)
                for i in range(len(status)):
                    assert status[i].usage == usage[i]
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_other_all_element_resonance():
    cmd_records = [
        [
            "sw_card 1 2",
            "choose 0",
            "TEST 2 card 2 cannot use",
            "skill 1 15 14 13",
            "TEST 2 card 2 cannot use",
            "TEST 3 card 1 target no 0",
            "card 0 1 12",
            "sw_char 0 12",
            "sw_char 1 11",
            "TEST 1 30 27 30 30 29 30",
            "TEST 4 p1c1 usage",
            "skill 1 10 9 8",
            "sw_char 0 7",
            "TEST 5 p0 team usage 2",
            "card 0 0 6",
            "TEST 5 p0 team usage 3 1",
            "skill 1 5 4 3",
            "end",
            "skill 1 15 14 13",
            "TEST 1 25 20 30 30 27 19",
            "skill 0 12 11 10",
            "card 3 0 9",
            "skill 1 8 7 6",
            "TEST 1 25 20 24 30 27 17",
            "card 0 0 5",
            "card 2 0 4",
            "end",
            "TEST 1 26 22 23 30 27 11",
            "card 0 0 15",
            "skill 2 14 13 12",
            "TEST 1 20 22 23 28 27 11",
            "skill 0 11 10 9",
            "skill 0 8 7 6",
            "sw_char 1 5",
            "skill 1 4 3 2",
            "TEST 2 card 0 cannot use",
            "end",
            "TEST 1 20 15 23 18 24 11",
            "skill 2 15 14 13",
            "sw_char 0 12",
            "sw_char 2 11",
            "skill 2 10 9 8 7",
            "card 4 0 6",
            "TEST 1 20 15 23 15 18 10",
            "sw_char 0 5",
            "card 0 0 4",
            "TEST 6 p0c1 charge 1",
            "sw_char 2 3",
            "end",
            "sw_char 0 15",
            "card 1 0 14",
            "skill 1 13 12 11",
            "TEST 1 13 15 22 15 1 10",
            "skill 0 10 9 8",
            "end",
            "end",
            "card 3 0 15",
            "TEST 1 14 15 23 12 0 7",
            "sw_char 2 14",
            "sw_char 0 13",
        ],
        [
            "sw_card 3 2",
            "choose 1",
            "card 0 0 15",
            "TEST 1 30 30 30 30 29 30",
            "sw_char 2 14",
            "sw_char 1 13",
            "card 0 0 12",
            "skill 1 11 10 9",
            "sw_char 2 8",
            "skill 1 7 6 5",
            "TEST 5 p0 team usage 2 1",
            "TEST 1 30 27 30 30 27 26",
            "card 0 0 4",
            "end",
            "TEST 1 25 27 30 30 27 21",
            "skill 0 15 14 13",
            "skill 2 12 11 10 9",
            "sw_char 0 8",
            "card 3 0 7",
            "skill 1 6 5 4",
            "sw_char 2 3",
            "end",
            "skill 1 15 14 13",
            "sw_char 1 12",
            "sw_char 0 11",
            "sw_char 1 10",
            "sw_char 0 9",
            "card 4 0 8",
            "TEST 7 p1 summon 2",
            "skill 1 7 6 5",
            "TEST 1 20 19 23 26 24 11",
            "end",
            "card 4 0 15",
            "sw_char 1 14",
            "sw_char 0 13",
            "sw_char 1 12",
            "sw_char 1 11",
            "end",
            "card 6 0 15",
            "TEST 6 p1c1 charge 2",
            "sw_char 0 14",
            "sw_char 1 13",
            "TEST 1 20 15 22 15 1 10",
            "card 5 0 12",
            "card 0 0 11",
            "skill 1 10 9 8",
            "choose 2",
            "sw_char 0 7",
            "end",
            "end",
            "sw_char 0 15",
            "card 5 0 14",
            "card 6 0 13",
            "TEST 6 p1c2 charge 1",
            "card 0 0 12",
            "card 2 0 12",
            "TEST 1 14 15 23 14 0 8",
            "TEST 3 card 3 target no 1",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Collei
        character:Fischl
        character:Xiangling
        Elemental Resonance: Shattering Ice*5
        Elemental Resonance: Soothing Water*5
        Elemental Resonance: Fervent Flames*5
        Elemental Resonance: High Voltage*5
        Elemental Resonance: Impetuous Winds*5
        Elemental Resonance: Sprawling Greenery*5
        """
    )
    for character in deck.characters:
        character.hp = 30
        character.max_hp = 30
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                hps = cmd.strip().split(" ")[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            elif test_id == 2:
                cmd = cmd.split()
                for req in match.requests:
                    if req.name == "UseCardRequest":
                        assert req.card_idx != int(cmd[3])
            elif test_id == 3:
                cmd = cmd.split()
                card_idx = int(cmd[3])
                target_idx = int(cmd[-1])
                for req in match.requests:
                    if req.name == "UseCardRequest":
                        if req.card_idx == card_idx:
                            for target in req.targets:
                                assert isinstance(target, ObjectPosition)
                                assert target.character_idx != target_idx
            elif test_id == 4:
                cmd = cmd.split()
                pid = int(cmd[2][1])
                cid = int(cmd[2][3])
                usage = [int(x) for x in cmd[4:]]
                status = match.player_tables[pid].characters[cid].status
                assert len(status) == len(usage)
            elif test_id == 5:
                cmd = cmd.split()
                pid = int(cmd[2][1])
                usage = [int(x) for x in cmd[5:]]
                status = match.player_tables[pid].team_status
                assert len(status) == len(usage)
                for u, s in zip(usage, status):
                    assert s.usage == u
            elif test_id == 6:
                cmd = cmd.split()
                pid = int(cmd[2][1])
                cid = int(cmd[2][3])
                charge = int(cmd[4])
                assert charge == match.player_tables[pid].characters[cid].charge
            elif test_id == 7:
                cmd = cmd.split()
                pid = int(cmd[2][1])
                usage = [int(x) for x in cmd[4:]]
                summons = match.player_tables[pid].summons
                assert len(summons) == len(usage)
                for u, s in zip(usage, summons):
                    assert s.usage == u
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_other_all_element_resonance_2():
    cmd_records = [
        [
            "sw_card 0 1 2",
            "choose 0",
            "skill 1 15 14 13",
            "skill 1 12 11 10",
            "end",
            "card 4 0 15",
            "card 3 0 14",
            "end",
            "sw_char 1 15",
            "card 4 0 14",
            "card 4 0 13",
            "card 2 0 12",
            "card 1 0 11",
            "skill 1 10 9 8",
            "TEST 1 29 30 30 21 24 24",
            "end",
        ],
        [
            "sw_card 3 2",
            "choose 0",
            "sw_char 1 15",
            "sw_char 2 14",
            "skill 1 13 12 11",
            "end",
            "end",
            "card 8 0 15",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Diluc
        character:AnemoMobMage
        character:Xingqiu
        Elemental Resonance: Shattering Ice*5
        Elemental Resonance: Soothing Water*5
        Elemental Resonance: Fervent Flames*5
        Elemental Resonance: High Voltage*5
        Elemental Resonance: Impetuous Winds*5
        Elemental Resonance: Sprawling Greenery*5
        """
    )
    for character in deck.characters:
        character.hp = 30
        character.max_hp = 30
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                hps = cmd.strip().split(" ")[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


if __name__ == "__main__":
    # test_woven_element()
    # test_enduring_rock()
    test_other_all_element_resonance()
    # test_other_all_element_resonance_2()



from typing import Dict, List, Literal

from ....action import CreateObjectAction, SwitchCharacterAction

from .....utils.class_registry import register_class
from ....event import RoundEndEventArguments
from ....consts import IconType, ObjectPositionType
from ....status.character_status.base import CharacterStatusBase
from ....match import Match
from ....struct import Cost, ObjectPosition
from ....object_base import EventCardBase
from .....utils.desc_registry import DescDictType


class FlickeringFourLeafSigilStatus_4_3(CharacterStatusBase):
    name: Literal["Flickering Four-Leaf Sigil"] = "Flickering Four-Leaf Sigil"
    version: Literal["4.3"] = "4.3"
    usage: int = 0
    max_usage: int = 0
    icon_type: IconType = IconType.SPECIAL

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Match
    ) -> List[SwitchCharacterAction]:
        """
        At the end of every round, switch to this character.
        """
        current_active = match.player_tables[
            self.position.player_idx
        ].active_character_idx
        if current_active == self.position.character_idx:
            # already active, do nothing
            return []
        return [
            SwitchCharacterAction(
                player_idx=self.position.player_idx,
                character_idx=self.position.character_idx,
            )
        ]


class FlickeringFourLeafSigil_4_3(EventCardBase):
    name: Literal["Flickering Four-Leaf Sigil"]
    version: Literal["4.3"] = "4.3"
    cost: Cost = Cost()

    def get_targets(self, match: Match) -> List[ObjectPosition]:
        # all alive characters
        ret: List[ObjectPosition] = []
        characters = match.player_tables[self.position.player_idx].characters
        for character in characters:
            if character.is_alive:
                ret.append(character.position)
        return ret

    def get_actions(
        self, target: ObjectPosition | None, match: Match
    ) -> List[CreateObjectAction]:
        assert target is not None
        return [
            CreateObjectAction(
                object_name=self.name,
                object_position=target.set_area(ObjectPositionType.CHARACTER_STATUS),
                object_arguments={},
            )
        ]


desc: Dict[str, DescDictType] = {
    "CHARACTER_STATUS/Flickering Four-Leaf Sigil": {
        "names": {"en-US": "Flickering Four-Leaf Sigil", "zh-CN": "浮烁的四叶印"},
        "descs": {
            "4.3": {
                "en-US": "At the End Phase of Every Round, you will switch to this character.",  # noqa: E501
                "zh-CN": "每个回合的结束阶段，我方都切换到此角色。",
            }
        },
    },
    "CARD/Flickering Four-Leaf Sigil": {
        "names": {"en-US": "Flickering Four-Leaf Sigil", "zh-CN": "浮烁的四叶印"},
        "descs": {
            "4.3": {
                "en-US": "Attached Four-Leaf Sigil to target character: At the End Phase of Every Round, you will switch to this character.",  # noqa: E501
                "zh-CN": "目标角色附属四叶印：每个回合的结束阶段，我方都切换到此角色。",
            }
        },
        "image_path": "cardface/Event_Event_Siyeyin.png",  # noqa: E501
        "id": 332027,
    },
}


register_class(FlickeringFourLeafSigil_4_3 | FlickeringFourLeafSigilStatus_4_3, desc)



from .struct import Cost, ObjectPosition
from ..utils import BaseModel
from typing import Literal, List
from .consts import DieColor, ElementalReactionType, ElementType, ObjectType
from .action import (
    ActionTypes,
    ActionBase,
    Actions,
    DrawCardAction,
    RestoreCardAction,
    RemoveCardAction,
    ChooseCharacterAction,
    CreateDiceAction,
    RemoveDiceAction,
    DeclareRoundEndAction,
    ActionEndAction,
    SkillEndAction,
    SwitchCharacterAction,
    MakeDamageAction,
    ChargeAction,
    CharacterDefeatedAction,
    CreateObjectAction,
    RemoveObjectAction,
    ChangeObjectUsageAction,
    MoveObjectAction,
    ConsumeArcaneLegendAction,
    UseCardAction,
    UseSkillAction,
    CharacterReviveAction,
)
from .modifiable_values import DamageValue, FinalDamageValue


class EventArgumentsBase(BaseModel):
    """
    Base class of event arguments. event arguments are generated when new
    Action is triggered by events. It is a superset of Action arguments, and
    will record nesessary information about the event.
    """

    type: Literal[ActionTypes.EMPTY] = ActionTypes.EMPTY
    action: ActionBase


class DrawCardEventArguments(EventArgumentsBase):
    """
    Event arguments for draw card event.
    """

    type: Literal[ActionTypes.DRAW_CARD] = ActionTypes.DRAW_CARD
    action: DrawCardAction
    hand_size: int
    max_hand_size: int


class RestoreCardEventArguments(EventArgumentsBase):
    """
    Event arguments for restore card event.
    """

    type: Literal[ActionTypes.RESTORE_CARD] = ActionTypes.RESTORE_CARD
    action: RestoreCardAction
    card_names: List[str]


class RemoveCardEventArguments(EventArgumentsBase):
    """
    Event arguments for remove card event.
    """

    type: Literal[ActionTypes.REMOVE_CARD] = ActionTypes.REMOVE_CARD
    action: RemoveCardAction
    card_name: str


class ChooseCharacterEventArguments(EventArgumentsBase):
    """
    Event arguments for choose character event.
    """

    type: Literal[ActionTypes.CHOOSE_CHARACTER] = ActionTypes.CHOOSE_CHARACTER
    action: ChooseCharacterAction
    original_character_idx: int


# 5


class CreateDiceEventArguments(EventArgumentsBase):
    """
    Event arguments for create dice event.

    Args:
        colors_generated (List[DieColor]): The colors of the dice
            generated.
        colors_over_maximum (List[DieColor]): The colors of the dice
            that are over the maximum number of dice and not obtained.
    """

    type: Literal[ActionTypes.CREATE_DICE] = ActionTypes.CREATE_DICE
    action: CreateDiceAction
    colors_generated: List[DieColor]
    colors_over_maximum: List[DieColor]


class RemoveDiceEventArguments(EventArgumentsBase):
    """
    Event arguments for remove dice event.

    Args:
        colors_removed (List[DieColor]): The colors of the dice.
    """

    type: Literal[ActionTypes.REMOVE_DICE] = ActionTypes.REMOVE_DICE
    action: RemoveDiceAction
    colors_removed: List[DieColor]


class GameStartEventArguments(EventArgumentsBase):
    """
    Event arguments for game start event. This event is triggered
    by the system when the game starts, so it does not have an action.

    Args:
        player_go_first (int): The index of the player who goes first.
    """

    type: Literal[ActionTypes.GAME_START] = ActionTypes.GAME_START
    action: ActionBase = ActionBase(type=ActionTypes.EMPTY)
    player_go_first: int


class RoundPrepareEventArguments(EventArgumentsBase):
    """
    Event arguments for round prepare event. This event is triggered
    by the system when a new round starts, so it does not have an action.

    Args:
        player_go_first (int): The index of the player who goes first.
        round (int): The number of the round.
        dice_colors (List[List[DiceColor]]): The colors of the dice of each
            player.
    """

    type: Literal[ActionTypes.ROUND_PREPARE] = ActionTypes.ROUND_PREPARE
    action: ActionBase = ActionBase(type=ActionTypes.EMPTY)
    player_go_first: int
    round: int
    dice_colors: List[List[DieColor]]


class DeclareRoundEndEventArguments(EventArgumentsBase):
    """
    Event arguments for declare round end event. This event is triggered
    by the system when a player declares the end of the round, so it does
    not have an action.
    """

    type: Literal[ActionTypes.DECLARE_ROUND_END] = ActionTypes.DECLARE_ROUND_END
    action: DeclareRoundEndAction


# 10


class ActionEndEventArguments(EventArgumentsBase):
    """
    Event arguments for action end event.
    """

    type: Literal[ActionTypes.ACTION_END] = ActionTypes.ACTION_END
    action: ActionEndAction
    do_combat_action: bool


class SwitchCharacterEventArguments(EventArgumentsBase):
    """
    Event arguments for switch character event.
    """

    type: Literal[ActionTypes.SWITCH_CHARACTER] = ActionTypes.SWITCH_CHARACTER
    action: SwitchCharacterAction
    last_active_character_idx: int


class ReceiveDamageEventArguments(EventArgumentsBase):
    """
    Event arguments for receive damage event. Some objects may trigger events
    before character defeated settlement start when received special
    damages, e.g. Seed of Skandha, Tenacity of the Millelith, Cryo Cicins.
    """

    type: Literal[ActionTypes.RECEIVE_DAMAGE] = ActionTypes.RECEIVE_DAMAGE
    action: MakeDamageAction
    original_damage: DamageValue
    final_damage: FinalDamageValue
    hp_before: int
    hp_after: int
    elemental_reaction: ElementalReactionType
    reacted_elements: List[ElementType]


class MakeDamageEventArguments(EventArgumentsBase):
    """
    Event arguments for make damage event. Some objects may trigger events
    before character defeated settlement start, e.g. reburn.
    """

    type: Literal[ActionTypes.MAKE_DAMAGE] = ActionTypes.MAKE_DAMAGE
    action: MakeDamageAction
    damages: List[ReceiveDamageEventArguments]


class AfterMakeDamageEventArguments(MakeDamageEventArguments):
    """
    Event arguments for make damage event, contains character defeated
    settlements.
    """

    type: Literal[ActionTypes.AFTER_MAKE_DAMAGE] = ActionTypes.AFTER_MAKE_DAMAGE
    action: MakeDamageAction
    damages: List[ReceiveDamageEventArguments]

    @staticmethod
    def from_make_damage_event_arguments(
        event_arguments: MakeDamageEventArguments,
    ) -> "AfterMakeDamageEventArguments":
        return AfterMakeDamageEventArguments(
            action=event_arguments.action,
            damages=event_arguments.damages,
        )


# 15


class ChargeEventArguments(EventArgumentsBase):
    """
    Event arguments for charge event.
    """

    type: Literal[ActionTypes.CHARGE] = ActionTypes.CHARGE
    action: ChargeAction
    charge_before: int
    charge_after: int


class UseSkillEventArguments(EventArgumentsBase):
    """
    Event arguments for use skill.
    """

    type: Literal[ActionTypes.USE_SKILL] = ActionTypes.USE_SKILL
    action: UseSkillAction


class SkillEndEventArguments(EventArgumentsBase):
    """
    Event arguments for skill end event.
    """

    type: Literal[ActionTypes.SKILL_END] = ActionTypes.SKILL_END
    action: SkillEndAction


class CharacterDefeatedEventArguments(EventArgumentsBase):
    """
    Event arguments for character defeated event.
    """

    type: Literal[ActionTypes.CHARACTER_DEFEATED] = ActionTypes.CHARACTER_DEFEATED
    action: CharacterDefeatedAction
    need_switch: bool


class CreateObjectEventArguments(EventArgumentsBase):
    """
    Event arguments for create object event.
    """

    type: Literal[ActionTypes.CREATE_OBJECT] = ActionTypes.CREATE_OBJECT
    action: CreateObjectAction
    create_result: Literal["NEW", "RENEW"]


# 20


class RemoveObjectEventArguments(EventArgumentsBase):
    """
    Event arguments for remove object event.
    """

    type: Literal[ActionTypes.REMOVE_OBJECT] = ActionTypes.REMOVE_OBJECT
    action: RemoveObjectAction
    object_name: str
    object_type: ObjectType


class ChangeObjectUsageEventArguments(EventArgumentsBase):
    """
    Event arguments for change object usage event.
    """

    type: Literal[ActionTypes.CHANGE_OBJECT_USAGE] = ActionTypes.CHANGE_OBJECT_USAGE
    action: ChangeObjectUsageAction
    object_name: str
    usage_before: int
    usage_after: int


class MoveObjectEventArguments(EventArgumentsBase):
    """
    Event arguments for move object event.
    """

    type: Literal[ActionTypes.MOVE_OBJECT] = ActionTypes.MOVE_OBJECT
    action: MoveObjectAction
    object_name: str


class ConsumeArcaneLegendEventArguments(EventArgumentsBase):
    """
    Event arguments for consume arcane legend event.
    """

    type: Literal[ActionTypes.CONSUME_ARCANE_LEGEND] = ActionTypes.CONSUME_ARCANE_LEGEND
    action: ConsumeArcaneLegendAction


class RoundEndEventArguments(EventArgumentsBase):
    """
    Event arguments for round end event. This event is triggered
    by the system when a round ends, so it does not have an action.

    Args:
        player_go_first (int): The index of the player who goes first.
        round (int): The number of the round.
    """

    type: Literal[ActionTypes.ROUND_END] = ActionTypes.ROUND_END
    action: ActionBase = ActionBase(type=ActionTypes.EMPTY)
    player_go_first: int
    round: int
    initial_card_draw: int


# 25


class PlayerActionStartEventArguments(EventArgumentsBase):
    """
    Event arguments for action start event.
    """

    type: Literal[ActionTypes.PLAYER_ACTION_START] = ActionTypes.PLAYER_ACTION_START
    action: ActionBase = ActionBase(type=ActionTypes.EMPTY)
    player_idx: int


class CharacterReviveEventArguments(EventArgumentsBase):
    """
    Event arguments for character revive event.
    """

    type: Literal[ActionTypes.CHARACTER_REVIVE] = ActionTypes.CHARACTER_REVIVE
    action: CharacterReviveAction


class UseCardEventArguments(EventArgumentsBase):
    """
    Event arguments for use card event.
    """

    type: Literal[ActionTypes.USE_CARD] = ActionTypes.USE_CARD
    action: UseCardAction
    card_cost: Cost
    card_name: str
    use_card_success: bool


EventArguments = (
    EventArgumentsBase
    | DrawCardEventArguments
    | RestoreCardEventArguments
    | RemoveCardEventArguments
    | ChooseCharacterEventArguments
    # 5
    | CreateDiceEventArguments
    | RemoveDiceEventArguments
    | GameStartEventArguments
    | RoundPrepareEventArguments
    | DeclareRoundEndEventArguments
    # 10
    | ActionEndEventArguments
    | SwitchCharacterEventArguments
    | ReceiveDamageEventArguments
    | MakeDamageEventArguments
    | AfterMakeDamageEventArguments
    # 15
    | ChargeEventArguments
    | UseSkillEventArguments
    | SkillEndEventArguments
    | CharacterDefeatedEventArguments
    | CreateObjectEventArguments
    # 20
    | RemoveObjectEventArguments
    | ChangeObjectUsageEventArguments
    | MoveObjectEventArguments
    | ConsumeArcaneLegendEventArguments
    | RoundEndEventArguments
    # 25
    | PlayerActionStartEventArguments
    | CharacterReviveEventArguments
    | UseCardEventArguments
)


class EventFrame(BaseModel):
    """
    Event frame is a frame of events, contains events with their arguments, the
    object lists that will be triggered by the event arguments, and the
    action lists that has triggered.

    When one action is done, it will generate a new event frame, with various
    number of events, and append it to the event frame list. When there are
    events and no processing objects and actions, the first event will
    trigger objects. When there are triggered objects, the first object
    position will be popped and trigger actions. When there are triggered
    actions, the first action will be popped and trigger events, then add
    a new event frame. If an event frame has no events, no triggered objects
    and no triggered actions, it will be removed from the list.
    """

    events: List[EventArguments]
    processing_event: EventArguments | None = None
    triggered_objects: List[ObjectPosition] = []
    triggered_actions: List[Actions] = []



from typing import Any, List, Literal

from .....utils.class_registry import register_class

from ....modifiable_values import DamageValue

from ....action import MakeDamageAction

from ....event import SkillEndEventArguments

from ....consts import DamageElementalType, DamageType, ObjectPositionType, WeaponType

from ....struct import Cost
from .base import RoundEffectWeaponBase


class AquilaFavonia_3_3(RoundEffectWeaponBase):
    name: Literal["Aquila Favonia"]
    cost: Cost = Cost(same_dice_number=3)
    version: Literal["3.3"] = "3.3"
    weapon_type: WeaponType = WeaponType.SWORD
    max_usage_per_round: int = 2

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[MakeDamageAction]:
        """
        If self is active and opposite use skill and has usage, heal 1 HP
        """
        if not self.position.check_position_valid(
            event.action.position,
            match,
            player_idx_same=False,
            source_area=ObjectPositionType.CHARACTER,
            target_area=ObjectPositionType.SKILL,
        ):
            # not equipped or not opponent use skill
            return []
        if (
            self.position.character_idx
            != match.player_tables[self.position.player_idx].active_character_idx
        ):
            # not active
            return []
        if self.usage == 0:
            # no usage
            return []
        self.usage -= 1
        character = match.player_tables[self.position.player_idx].characters[
            self.position.character_idx
        ]
        return [
            MakeDamageAction(
                damage_value_list=[
                    DamageValue(
                        position=self.position,
                        damage_type=DamageType.HEAL,
                        target_position=character.position,
                        damage=-1,
                        damage_elemental_type=DamageElementalType.HEAL,
                        cost=self.cost.copy(),
                    )
                ],
            )
        ]


register_class(AquilaFavonia_3_3)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...summon.base import AttackerSummonBase

from ...action import Actions
from ...struct import Cost

from ...consts import (
    DamageElementalType,
    DieColor,
    ElementType,
    FactionType,
    WeaponType,
)
from ..character_base import (
    AOESkillBase,
    ElementalBurstBase,
    ElementalSkillBase,
    PhysicalNormalAttackBase,
    CharacterBase,
    SkillTalent,
)


# Summons


class ClusterbloomArrow_3_6(AttackerSummonBase):
    name: Literal["Clusterbloom Arrow"] = "Clusterbloom Arrow"
    version: Literal["3.6"] = "3.6"
    usage: int = 1
    max_usage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.DENDRO
    damage: int = 1
    renew_type: Literal["ADD"] = "ADD"


# Skills


class VijnanaPhalaMine(ElementalSkillBase):
    name: Literal["Vijnana-Phala Mine"] = "Vijnana-Phala Mine"
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.DENDRO
    cost: Cost = Cost(elemental_dice_color=DieColor.DENDRO, elemental_dice_number=3)

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create object
        """
        return super().get_actions(
            match,
            [
                self.create_character_status("Vijnana Suffusion"),
            ],
        )


class FashionersTanglevineShaft(ElementalBurstBase, AOESkillBase):
    name: Literal["Fashioner's Tanglevine Shaft"] = "Fashioner's Tanglevine Shaft"
    damage: int = 4
    damage_type: DamageElementalType = DamageElementalType.DENDRO
    back_damage: int = 1
    cost: Cost = Cost(
        elemental_dice_color=DieColor.DENDRO, elemental_dice_number=3, charge=2
    )


# Talents


class KeenSight_3_6(SkillTalent):
    name: Literal["Keen Sight"]
    version: Literal["3.6"] = "3.6"
    character_name: Literal["Tighnari"] = "Tighnari"
    cost: Cost = Cost(
        elemental_dice_color=DieColor.DENDRO,
        elemental_dice_number=4,
    )
    skill: Literal["Vijnana-Phala Mine"] = "Vijnana-Phala Mine"


# character base


class Tighnari_3_6(CharacterBase):
    name: Literal["Tighnari"]
    version: Literal["3.6"] = "3.6"
    element: ElementType = ElementType.DENDRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase | VijnanaPhalaMine | FashionersTanglevineShaft
    ] = []
    faction: List[FactionType] = [FactionType.SUMERU]
    weapon_type: WeaponType = WeaponType.BOW

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name="Khanda Barrier-Buster",
                cost=PhysicalNormalAttackBase.get_cost(self.element),
            ),
            VijnanaPhalaMine(),
            FashionersTanglevineShaft(),
        ]


register_class(Tighnari_3_6 | KeenSight_3_6 | ClusterbloomArrow_3_6)



import numpy as np
from pydantic import PrivateAttr
from .agent_base import AgentBase
from ..server.match import Match
from ..server.interaction import (
    Responses,
    SwitchCardRequest,
    SwitchCardResponse,
    ChooseCharacterRequest,
    ChooseCharacterResponse,
    RerollDiceRequest,
    RerollDiceResponse,
    DeclareRoundEndResponse,
    ElementalTuningRequest,
    ElementalTuningResponse,
    SwitchCharacterRequest,
    SwitchCharacterResponse,
    UseSkillRequest,
    UseSkillResponse,
    UseCardRequest,
    UseCardResponse,
)
from ..server.consts import DieColor


class RandomAgent(AgentBase):
    """
    Agent that randomly choose one type of requests, randomly choose one of
    selected type, and randomly choose one of available options.
    """

    random_seed: int | None = None
    _random_state: np.random.RandomState = PrivateAttr(np.random.RandomState())
    random_state_set: bool = False

    def json(self, *argv, **kwargs):
        raise NotImplementedError("RandomAgent does not support json now")

    def random(self) -> float:
        """
        Return a random float between 0 and 1.
        """
        if not self.random_state_set:
            self._random_state = np.random.RandomState()
            if self.random_seed is None:
                self.random_seed = np.random.randint(0, 2**31 - 1)
            self._random_state.seed(self.random_seed)
            self.random_state_set = True
        return self._random_state.rand()

    def generate_response(self, match: Match) -> Responses | None:
        req_names = list(
            set([x.name for x in match.requests if x.player_idx == self.player_idx])
        )
        req_names.sort()
        if len(req_names) == 0:
            return None
        req_name = req_names[int(self.random() * len(req_names))]
        reqs = [
            x
            for x in match.requests
            if x.player_idx == self.player_idx and x.name == req_name
        ]
        req = reqs[int(self.random() * len(reqs))]

        if req.name == "SwitchCardRequest":
            return self.resp_switch_card(req)
        elif req.name == "ChooseCharacterRequest":
            return self.resp_choose_character(req)
        elif req.name == "RerollDiceRequest":
            return self.resp_reroll_dice(req)
        elif req.name == "DeclareRoundEndRequest":
            return DeclareRoundEndResponse(request=req)
        elif req.name == "ElementalTuningRequest":
            return self.resp_elemental_tuning(req)
        elif req.name == "SwitchCharacterRequest":
            return self.resp_switch_character(req)
        elif req.name == "UseSkillRequest":
            return self.resp_use_skill(req)
        elif req.name == "UseCardRequest":
            return self.resp_use_card(req)
        else:
            raise ValueError(f"Unknown request name: {req.name}")

    def resp_switch_card(self, req: SwitchCardRequest) -> SwitchCardResponse:
        """
        Randomly choose a subset of cards.
        """
        card_idxs = []
        for i in range(len(req.card_names)):
            if self.random() < 0.5:
                card_idxs.append(i)
        return SwitchCardResponse(request=req, card_idxs=card_idxs)

    def resp_choose_character(
        self, req: ChooseCharacterRequest
    ) -> ChooseCharacterResponse:
        """
        Randomly choose a character.
        """
        return ChooseCharacterResponse(
            request=req,
            character_idx=req.available_character_idxs[
                int(self.random() * len(req.available_character_idxs))
            ],
        )

    def resp_reroll_dice(self, req: RerollDiceRequest) -> RerollDiceResponse:
        """
        Randomly choose a subset of dice.
        """
        reroll_dice_idxs = []
        for i in range(len(req.colors)):
            if self.random() < 0.5:
                reroll_dice_idxs.append(i)
        return RerollDiceResponse(request=req, reroll_dice_idxs=reroll_dice_idxs)

    def resp_elemental_tuning(
        self, req: ElementalTuningRequest
    ) -> ElementalTuningResponse:
        """
        Randomly choose a subset of dice.
        """
        return ElementalTuningResponse(
            request=req,
            dice_idx=req.dice_idxs[int(self.random() * len(req.dice_idxs))],
            card_idx=req.card_idxs[int(self.random() * len(req.card_idxs))],
        )

    def resp_switch_character(
        self, req: SwitchCharacterRequest
    ) -> SwitchCharacterResponse:
        """
        Randomly choose a character.
        """
        assert req.cost.any_dice_number <= 1
        idxs = []
        if req.cost.any_dice_number == 1:
            idxs = [int(self.random() * len(req.dice_colors))]
        return SwitchCharacterResponse(
            request=req,
            dice_idxs=idxs,
        )

    def resp_use_skill(self, req: UseSkillRequest) -> UseSkillResponse:
        """
        Randomly choose dice to use skill.
        """
        cost = req.cost
        ele_dice_idxs = [
            num
            for num, color in enumerate(req.dice_colors)
            if color == cost.elemental_dice_color or color == DieColor.OMNI
        ]
        other_dice_idxs = [
            x for x in range(len(req.dice_colors)) if x not in ele_dice_idxs
        ]
        selected = []
        for _ in range(cost.elemental_dice_number):
            if len(ele_dice_idxs) > 0:
                idx = int(self.random() * len(ele_dice_idxs))
                selected.append(ele_dice_idxs.pop(idx))
            else:
                raise ValueError("Not enough elemental dice")
        other_dice_idxs += ele_dice_idxs
        for _ in range(cost.any_dice_number):
            if len(other_dice_idxs) > 0:
                idx = int(self.random() * len(other_dice_idxs))
                selected.append(other_dice_idxs.pop(idx))
            else:
                raise ValueError("Not enough dice")
        selected.sort()

        return UseSkillResponse(request=req, dice_idxs=selected)

    def resp_use_card(self, req: UseCardRequest) -> UseCardResponse:
        """
        Randomly choose dice to use card.
        """
        cost = req.cost
        ele_dice_idxs = [
            num
            for num, color in enumerate(req.dice_colors)
            if color == cost.elemental_dice_color or color == DieColor.OMNI
        ]
        other_dice_idxs = [
            x for x in range(len(req.dice_colors)) if x not in ele_dice_idxs
        ]
        selected = []
        for _ in range(cost.elemental_dice_number):
            if len(ele_dice_idxs) > 0:
                idx = int(self.random() * len(ele_dice_idxs))
                selected.append(ele_dice_idxs.pop(idx))
            else:
                raise ValueError("Not enough elemental dice")
        if cost.same_dice_number > 0:
            color_map = {}
            for color in req.dice_colors:
                if color not in color_map:
                    color_map[color] = 0
                color_map[color] += 1
            keys = list(color_map.keys())
            keys.sort()
            if DieColor.OMNI not in color_map:
                color_map[DieColor.OMNI] = 0
            available_keys = []
            for k in keys:
                v = color_map[k]
                if v >= cost.same_dice_number:
                    available_keys.append(k)
                elif (
                    k != DieColor.OMNI
                    and v + color_map[DieColor.OMNI] >= cost.same_dice_number
                ):
                    available_keys.append(k)
            if len(available_keys) > 0:
                k = available_keys[int(self.random() * len(available_keys))]
                other_dice_idxs = []
                ele_dice_idxs = []
                for i, color in enumerate(req.dice_colors):
                    if color == k or color == DieColor.OMNI:
                        if len(selected) < cost.same_dice_number:
                            selected.append(i)
                        else:
                            other_dice_idxs.append(i)
                    else:
                        other_dice_idxs.append(i)
        other_dice_idxs += ele_dice_idxs
        for _ in range(cost.any_dice_number):
            if len(other_dice_idxs) > 0:
                idx = int(self.random() * len(other_dice_idxs))
                selected.append(other_dice_idxs.pop(idx))
            else:
                raise ValueError("Not enough dice")
        selected.sort()

        target = None
        if len(req.targets):
            target = req.targets[int(self.random() * len(req.targets))]

        return UseCardResponse(
            request=req,
            dice_idxs=selected,
            target=target,
        )



from typing import Dict, List, Literal

from .....utils.class_registry import register_class

from ....status.team_status.base import RoundTeamStatus

from ....consts import DieColor, IconType, ObjectPositionType, ObjectType
from ....action import CreateDiceAction, CreateObjectAction, RemoveObjectAction
from ....match import Match
from ....event import RemoveObjectEventArguments, RoundPrepareEventArguments
from ....struct import Cost, ObjectPosition
from ....object_base import EventCardBase
from .....utils.desc_registry import DescDictType


class TheBoarPrincessStatus_4_3(RoundTeamStatus):
    name: Literal["The Boar Princess"] = "The Boar Princess"
    version: Literal["4.3"] = "4.3"
    usage: int = 2
    max_usage: int = 2
    icon_type: IconType = IconType.BUFF

    def event_handler_REMOVE_OBJECT(
        self, event: RemoveObjectEventArguments, match: Match
    ) -> List[CreateDiceAction | RemoveObjectAction]:
        """
        When self equipment removed from character, create one omni die
        """
        position = event.action.object_position
        if position.player_idx != self.position.player_idx:
            # not self
            return []
        if position.area != ObjectPositionType.CHARACTER:
            # not remove character objects, i.e. equipment
            return []
        if event.object_type == ObjectType.TALENT:
            # is talent, it should be removed because of character defeated.
            character = match.player_tables[position.player_idx].characters[
                position.character_idx
            ]
            if character.is_alive:
                # character is still alive, it should be because of
                # talent overwritten.
                return []
        assert self.usage > 0
        self.usage -= 1
        return [
            CreateDiceAction(
                player_idx=self.position.player_idx, number=1, color=DieColor.OMNI
            )
        ] + self.check_should_remove()

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Match
    ) -> List[RemoveObjectAction]:
        self.usage = 0
        return self.check_should_remove()


class TheBoarPrincess_4_3(EventCardBase):
    name: Literal["The Boar Princess"]
    version: Literal["4.3"] = "4.3"
    cost: Cost = Cost()

    def get_targets(self, match: Match) -> List[ObjectPosition]:
        # no target
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Match
    ) -> List[CreateObjectAction]:
        assert target is None
        return [
            CreateObjectAction(
                object_name=self.name,
                object_position=ObjectPosition(
                    player_idx=self.position.player_idx,
                    area=ObjectPositionType.TEAM_STATUS,
                    id=-1,
                ),
                object_arguments={},
            )
        ]


desc: Dict[str, DescDictType] = {
    "TEAM_STATUS/The Boar Princess": {
        "names": {"en-US": "The Boar Princess", "zh-CN": "野猪公主"},
        "descs": {
            "4.3": {
                "en-US": "Each time you discard an Equipment Card from one of your characters during this Round: Gain 1 Omni Element. (Max 2)\n(This effect can be triggered by the loss of an Equipment Card from a character falling or from having their Weapon or Artifact overwritten.)",  # noqa: E501
                "zh-CN": "本回合中，我方每有一张装备在角色身上的「装备牌」被弃置时：获得1个万能元素。（最多获得2个）\n（角色被击倒时弃置装备牌，或者覆盖装备「武器」或「圣遗物」，都可以触发此效果）",  # noqa: E501
            }
        },
    },
    "CARD/The Boar Princess": {
        "names": {"en-US": "The Boar Princess", "zh-CN": "野猪公主"},
        "descs": {
            "4.3": {
                "en-US": "Each time you discard an Equipment Card from one of your characters during this Round: Gain 1 Omni Element. (Max 2)\n(This effect can be triggered by the loss of an Equipment Card from a character falling or from having their Weapon or Artifact overwritten.)",  # noqa: E501
                "zh-CN": "本回合中，我方每有一张装备在角色身上的「装备牌」被弃置时：获得1个万能元素。（最多获得2个）\n（角色被击倒时弃置装备牌，或者覆盖装备「武器」或「圣遗物」，都可以触发此效果）",  # noqa: E501
            }
        },
        "image_path": "cardface/Event_Event_Yezhu.png",  # noqa: E501
        "id": 332025,
    },
}


register_class(TheBoarPrincess_4_3 | TheBoarPrincessStatus_4_3, desc)



from typing import Literal, List, Any

from ....utils.class_registry import register_base_class

from ...event import MoveObjectEventArguments, RoundPrepareEventArguments
from ...object_base import CardBase
from ...consts import IconType, ObjectType, ObjectPositionType
from ...struct import Cost
from ...action import Actions, RemoveObjectAction, MoveObjectAction
from ...struct import ObjectPosition


class SupportBase(CardBase):
    """
    Supports are cards that can be played on the table. Based on its position,
    it will have different actions. When in hand, it performs like a event
    card that have costs, is_valid, and get_actions. When on the table, its
    event triggers will work and do supports.
    """

    name: str
    version: str
    cost: Cost
    usage: int
    type: Literal[ObjectType.SUPPORT] = ObjectType.SUPPORT
    remove_when_used: bool = False

    # icon type is used to show the icon on the summon top right.
    icon_type: Literal[
        IconType.SHIELD,
        IconType.BARRIER,
        IconType.TIMESTATE,
        IconType.COUNTER,
        IconType.NONE,
    ]
    # when status icon type is not none, it will show in team status area
    status_icon_type: Literal[IconType.NONE] = IconType.NONE

    def check_should_remove(self) -> List[RemoveObjectAction]:
        """
        Check if the support should be removed.
        when usage has changed, call this function to check if the support
        should be removed.
        """
        if self.position.area != ObjectPositionType.SUPPORT:  # pragma: no cover
            return []
        if self.usage <= 0:
            return [
                RemoveObjectAction(
                    object_position=self.position,
                )
            ]
        return []

    def play(self, match: Any) -> List[Actions]:
        """
        Triggered when the support is played from hand to support area.
        """
        return []

    def is_valid(self, match: Any) -> bool:
        """
        If it is not in hand, cannot use.
        """
        return self.position.area == ObjectPositionType.HAND

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        """
        When support field is full, should choose one target to remove.
        Otherwise, return empty.
        """
        max_support_number = match.config.max_support_number
        supports = match.player_tables[self.position.player_idx].supports
        assert max_support_number >= len(supports)
        if len(supports) == max_support_number:
            # choose one support to remove
            ret: List[ObjectPosition] = []
            for support in supports:
                ret.append(support.position)
            return ret
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[RemoveObjectAction | MoveObjectAction]:
        """
        Act the support. will place it into support area.
        """
        ret: List[RemoveObjectAction | MoveObjectAction] = []
        max_support_number = match.config.max_support_number
        supports = match.player_tables[self.position.player_idx].supports
        assert max_support_number >= len(supports)
        if len(supports) == max_support_number:
            assert target is not None
            ret.append(RemoveObjectAction(object_position=target))
        ret.append(
            MoveObjectAction(
                object_position=self.position,
                target_position=ObjectPosition(
                    player_idx=self.position.player_idx,
                    area=ObjectPositionType.SUPPORT,
                    id=self.position.id,
                ),
            )
        )
        return ret

    def event_handler_MOVE_OBJECT(
        self, event: MoveObjectEventArguments, match: Any
    ) -> List[Actions]:
        """
        When this support is moved from hand to support, it is considered
        as played, and will call `self.play`.
        """
        if (
            event.action.object_position.id == self.id
            and event.action.object_position.area == ObjectPositionType.HAND
            and event.action.target_position.area == ObjectPositionType.SUPPORT
        ):
            # this artifact equipped from hand to character
            return self.play(match)
        return []


register_base_class(SupportBase)


class RoundEffectSupportBase(SupportBase):
    """
    Supports that has round effects. Refresh their usage when played and
    at round preparing stage.
    Instead of setting usage, set max_usage_per_round.
    """

    name: str
    version: str
    cost: Cost
    max_usage_per_round: int

    usage: int = 0

    icon_type: Literal[IconType.NONE] = IconType.NONE

    def play(self, match: Any) -> List[Actions]:
        self.usage = self.max_usage_per_round
        return []

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[Actions]:
        self.usage = self.max_usage_per_round
        return []


class UsageWithRoundRestrictionSupportBase(SupportBase):
    """
    Supports that restricts maximum usage per round, and also has total
    maximum usage. e.g. Liu Su has 1 usage per round and 2 total usage.
    """

    name: str
    version: str
    cost: Cost
    usage: int = 2
    usage_this_round: int = 0
    max_usage_one_round: int

    icon_type: Literal[IconType.TIMESTATE] = IconType.TIMESTATE

    def play(self, match: Any) -> List[Actions]:
        self.usage_this_round = self.max_usage_one_round
        return super().play(match)

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[Actions]:
        self.usage_this_round = self.max_usage_one_round
        return []

    def has_usage(self) -> bool:
        return self.usage > 0 and self.usage_this_round > 0

    def use(self) -> None:
        assert self.has_usage()
        self.usage -= 1
        self.usage_this_round -= 1


class LimitedEffectSupportBase(SupportBase):
    """
    Supports that has limited effect during the game. e.g. Chef Mao in 4.1
    will draw 1 food card when triggered only once.

    Overwrite the _limited_action function to do the limited action, and call
    do_limited_action when the condition is satisfied.
    """

    name: str
    version: str
    cost: Cost
    limited_usage: int

    def _limited_action(self, match: Any) -> List[Actions]:
        """
        return limited action.
        """
        raise NotImplementedError()

    def do_limited_action(self, match: Any) -> List[Actions]:
        """
        do the limited action when has usage and is in support area.
        """
        if self.limited_usage <= 0 or self.position.area != ObjectPositionType.SUPPORT:
            return []
        self.limited_usage -= 1
        return self._limited_action(match)



"""
Serve a room server on localhost:7999. Note you should wrap the server with
`if __name__ == '__main__':` to avoid error.
"""


from src.lpsim.network.http_room_server import HTTPRoomServer


if __name__ == "__main__":
    room_server = HTTPRoomServer(
        max_rooms=3,
        port=7999,
        room_port_range=[8001, 8010],
        room_timeout=300,
        admin_password="foobar",
        run_args={},
    )
    room_server.run()
    print("server closed")
